import torch.nn as nn
import torch
import numpy as np
import torchvision.transforms.functional as F
from torchvision import transforms, datasets
from torchvision import datasets, models, transforms

torch.manual_seed(1)
np.random.seed(0)

class Binarize(nn.Module):
    def __init__(self):
        super(Binarize, self).__init__()
    def forward(self, x):
        return torch.where(x > 0, 1, 0).float()

class SquarePad:
	def __call__(self, image):
		w, h = image.size
		max_wh = np.max([w, h])
		hp = int((max_wh - w) / 2)
		vp = int((max_wh - h) / 2)
		padding = (hp, vp, hp, vp)
		return F.pad(image, padding, 0, 'constant')

class LookingModel(nn.Module):
    """
    Class definition for the Looking Model.
    """
    def __init__(self, input_size, p_dropout=0.2, output_size=1, linear_size=256, num_stage=3):
        """[summary]

        Args:
            input_size (int): Input size for the model. If the whole pose needs to be used, the value should be 51.
            p_dropout (float, optional): Dropout rate in the linear blocks. Defaults to 0.2.
            output_size (int, optional): Output number of nodes. Defaults to 1.
            linear_size (int, optional): Size of the fully connected layers in the Linear blocks. Defaults to 256.
            num_stage (int, optional): Number of stages to use in the Linear Block. Defaults to 3.
            bce (bool, optional): Make use of the BCE Loss. Defaults to False.
        """
        super(LookingModel, self).__init__()

        self.input_size = input_size
        self.output_size = output_size
        self.linear_size = linear_size
        self.p_dropout = p_dropout
        self.num_stage = num_stage
        self.relu = nn.ReLU(inplace=True)
        self.dropout = nn.Dropout(self.p_dropout)

        # process input to linear size
        self.w1 = nn.Linear(self.input_size, self.linear_size)
        self.batch_norm1 = nn.BatchNorm1d(self.linear_size)

        self.linear_stages = []
        for _ in range(num_stage):
            self.linear_stages.append(Linear(self.linear_size, self.p_dropout))
        self.linear_stages = nn.ModuleList(self.linear_stages)

        # post processing
        self.w2 = nn.Linear(self.linear_size, self.output_size)
        self.sigmoid = nn.Sigmoid()

    def forward_first_stage(self, x):
        # pre-processing
        y = self.w1(x)
        y = self.batch_norm1(y)
        y = self.relu(y)
        y = self.dropout(y)
        return y
    
    def forward_second_stage(self, y):
        for i in range(self.num_stage):
            y = self.linear_stages[i](y)
        #y = self.binarize(y)
        y = self.w2(y)
        #y = self.binarize(y)
        y = self.sigmoid(y)
        return y

    def forward(self, x):
        # pre-processing
        y = self.w1(x)
        y = self.batch_norm1(y)
        y = self.relu(y)
        y = self.dropout(y)
        # linear layers
        for i in range(self.num_stage):
            y = self.linear_stages[i](y)
        #y = self.binarize(y)
        y = self.w2(y)
        #y = self.binarize(y)
        y = self.sigmoid(y)
        return y


class Linear(nn.Module):
    """
    Class definition of the Linear block
    """
    def __init__(self, linear_size=256, p_dropout=0.2):
        """
        Args:
            linear_size (int, optional): Size of the FC layers inside the block. Defaults to 256.
            p_dropout (float, optional): Dropout rate. Defaults to 0.2.
        """
        super(Linear, self).__init__()
        ###

        self.linear_size = linear_size
        self.p_dropout = p_dropout

        ###

        self.relu = nn.ReLU(True)
        self.dropout = nn.Dropout(self.p_dropout)

        ###
        self.l1 = nn.Linear(self.linear_size, self.linear_size)
        self.bn1 = nn.BatchNorm1d(self.linear_size)

        self.l2 = nn.Linear(self.linear_size, self.linear_size)
        self.bn2 = nn.BatchNorm1d(self.linear_size)
        
    def forward(self, x):
        # stage I

        y = self.l1(x)
        y = self.bn1(y)
        y = self.relu(y)
        y = self.dropout(y)

        # stage II

        
        y = self.l2(y)
        y = self.bn2(y)
        y = self.relu(y)
        y = self.dropout(y)

        # concatenation

        out = x+y

        return out

class AlexNet_head(nn.Module):
    """
    Class definition of the AlexNet crops model (Rasouli et al.).
    """
    def __init__(self, device, fine_tune=True):
        """
        Args:
            device (torch device): PyTorch device 
            fine_tune (bool, optional): Use the finetuned model. Defaults to True.
        """
        super(AlexNet_head, self).__init__()
        net = models.alexnet(pretrained=fine_tune).to(device)
        net.classifier  = nn.Sequential(
                    nn.Dropout(),
                    nn.Linear(256 * 6 * 6, 4096),
                    nn.ReLU(inplace=True),
                    nn.Dropout(),
                    nn.Linear(4096, 4096),
                    nn.ReLU(inplace=True),
                    nn.Linear(4096, 1),
                    nn.Sigmoid()
        ).to(device)
        if fine_tune:
            for param in net.parameters():
                param.requires_grad = False
            for param in net.classifier.parameters():
                param.requires_grad = True
            print('Fine Tune : ', fine_tune)
        self.net = net
    def forward(self, x):
        return self.net(x)

class ResNet18_head(nn.Module):
    def __init__(self, device, fine_tune=True):
        """
        Args:
            device (torch device): PyTorch device 
            fine_tune (bool, optional): Use the finetuned model. Defaults to True.
        """
        super(ResNet18_head, self).__init__()
        self.net = models.resnet18(pretrained=True)
        self.net.fc  = nn.Sequential(
            nn.Linear(in_features=512, out_features=1, bias=True),
            nn.Sigmoid()
        ).to(device)
        if fine_tune:
            for param in self.net.parameters():
                param.requires_grad = False
            for param in self.net.fc.parameters():
                param.requires_grad = True
        print('Fine Tune : ', fine_tune)

    def forward(self, x):
        return self.net(x)

class ResNet50_head(nn.Module):
    def __init__(self, device, fine_tune=True):
        """
        Args:
            device (torch device): PyTorch device 
            fine_tune (bool, optional): Use the finetuned model. Defaults to True.
        """
        super(ResNet50_head, self).__init__()
        net = models.resnext50_32x4d(pretrained=True)
        net.fc  = nn.Sequential(
            nn.Linear(in_features=2048, out_features=1, bias=True),
            nn.Sigmoid()
        ).to(device)
        if fine_tune:
            for param in net.parameters():
                param.requires_grad = False
            for param in net.fc.parameters():
                param.requires_grad = True
        self.net = net
        print('Fine Tune : ', fine_tune)

    def forward(self, x):
        return self.net(x)

class LookingNet_late_fusion_18(nn.Module):
    """
        Class definition for the combined Looking Model. Late fusion architecture with Resnet18 backbone. 
    """
    def __init__(self, PATH, PATH_look, device, fine_tune=True):
        """
        Args:
            PATH (str): Path to the pretrained ResNet18 heads model. Applicable only if fine-tune is enabled
            PATH_look (str): Path to the pretrained Looking joints model. Applicable only if fine-tune is enabled
            device (PyTorch device): PyTorch device
            fine_tune (bool, optional): Enable fine tune. Defaults to True.
        """
        super(LookingNet_late_fusion_18, self).__init__()
        self.backbone = ResNet18_head(device)
        if fine_tune:
            self.backbone.load_state_dict(torch.load(PATH))
            for m in self.backbone.net.parameters():
                m.requires_grad = False
        self.backbone.eval()

        self.looking_model = LookingModel(51)
        if fine_tune:
            self.looking_model.load_state_dict(torch.load(PATH_look, map_location=torch.device(device)))
            for m in self.looking_model.parameters():
                m.requires_grad = False
            self.looking_model.eval()

        self.relu = nn.ReLU(inplace=True)

        self.encoder_head = nn.Sequential(
            nn.Linear(512, 256, bias=False),
            nn.BatchNorm1d(256),
            nn.Dropout(0.4),
            nn.ReLU(inplace=True)
        )

        self.final = nn.Sequential(
            nn.Linear(512, 1),
            nn.Sigmoid()
        )


    def forward(self, x):
        head, keypoint = x
        activation = {}
        def get_activation(name):
            def hook(model, input, output):
                activation[name] = output.detach().squeeze()
            return hook
        def get_activation2(name):
            def hook(model, input, output):
                activation[name] = output.detach()
            return hook
        self.looking_model.linear_stages[2].bn2.register_forward_hook(get_activation('look'))
        self.backbone.net.avgpool.register_forward_hook(get_activation('avgpool'))
        _ = self.backbone(head)
        _ = self.looking_model(keypoint)

        layer_look = activation["look"]
        layer_resnet = activation["avgpool"]

        out_final = torch.cat((self.encoder_head(layer_resnet), self.relu(layer_look)), 1).type(torch.float)
        return self.final(out_final)


class LookingNet_late_fusion_50(nn.Module):
    """
        Class definition for the combined Looking Model. Late fusion architecture with ResNext50 backbone. 
    """
    def __init__(self, PATH, PATH_look, device, fine_tune=True):
        """
        Args:
            PATH (str): Path to the pretrained ResNext50 heads model. Applicable only if fine-tune is enabled
            PATH_look (str): Path to the pretrained Looking joints model. Applicable only if fine-tune is enabled
            device (PyTorch device): PyTorch device
            fine_tune (bool, optional): Enable fine tune. Defaults to True.
        """
        super(LookingNet_late_fusion_50, self).__init__()
        self.backbone = ResNet50_head(device)
        if fine_tune:
            self.backbone.load_state_dict(torch.load(PATH))
            for m in self.backbone.net.parameters():
                m.requires_grad = False
            self.backbone.eval()

        self.looking_model = LookingModel(51)
        if fine_tune:
            self.looking_model.load_state_dict(torch.load(PATH_look, map_location=torch.device(device)))
            for m in self.looking_model.parameters():
                m.requires_grad = False
            self.looking_model.eval()

        self.relu = nn.ReLU(inplace=True)

        self.encoder_head = nn.Sequential(
            nn.Linear(2048, 256, bias=False),
            nn.BatchNorm1d(256),
            nn.Dropout(0.4),
            nn.ReLU(inplace=True)
        )

        self.final = nn.Sequential(
            nn.Linear(512, 1),
            nn.Sigmoid()
        )


    def forward(self, x):
        head, keypoint = x
        activation = {}
        def get_activation(name):
            def hook(model, input, output):
                activation[name] = output.detach().squeeze()
            return hook
        def get_activation2(name):
            def hook(model, input, output):
                activation[name] = output.detach()
            return hook
        self.looking_model.linear_stages[2].bn2.register_forward_hook(get_activation('look'))
        self.backbone.net.avgpool.register_forward_hook(get_activation('avgpool'))
        _ = self.backbone(head)
        _ = self.looking_model(keypoint)

        layer_look = activation["look"]
        layer_resnet = activation["avgpool"]

        out_final = torch.cat((self.encoder_head(layer_resnet), self.relu(layer_look)), 1).type(torch.float)
        return self.final(out_final)

class LookingNet_early_fusion_eyes(nn.Module):
    """
        Class definition for the combined Looking Model. Early fusion architecture with the eyes crops. 
    """
    def __init__(self, PATH_look, device, fine_tune=True):
        """
        Args:
            PATH_look (str): Path to the pretrained Looking joints model. Applicable only if fine-tune is enabled
            device (PyTorch device): PyTorch device
            fine_tune (bool, optional): Enable fine tune. Defaults to True.
        """
        super(LookingNet_early_fusion_eyes, self).__init__()
        self.looking_model = LookingModel(51)
        print("Fine tune : " , fine_tune)
        if fine_tune:
            self.looking_model.load_state_dict(torch.load(PATH_look, map_location=torch.device(device)))
            for m in self.looking_model.parameters():
                m.requires_grad = False
            self.looking_model.eval()


        self.encoder_eyes = nn.Sequential(
            nn.Flatten(),
            nn.Linear(900, 256, bias=False),
            nn.BatchNorm1d(256),
            nn.Dropout(0.4),
            nn.ReLU(inplace=True)
        )

        self.looking_module = Looking_early_module()
    def forward(self, x):
        
        eyes, keypoint = x
        encoded_eyes = self.encoder_eyes(eyes)

        output_first_stage = self.looking_model.forward_first_stage(keypoint)
        y = self.looking_module.forward(output_first_stage+encoded_eyes)
        return y

class Looking_early_module(nn.Module):
    def __init__(self, p_dropout=0.2, output_size=1, linear_size=256, num_stage=3):
        super(Looking_early_module, self).__init__()
        self.p_dropout = p_dropout
        self.linear_stages = []
        self.linear_size = linear_size
        self.output_size = output_size
        self.num_stage = num_stage

        for _ in range(num_stage):
            self.linear_stages.append(Linear(self.linear_size, self.p_dropout))
        self.linear_stages = nn.ModuleList(self.linear_stages)

        self.w2 = nn.Linear(self.linear_size, self.output_size)
        self.sigmoid = nn.Sigmoid()
    
    def forward(self, y):
        for i in range(self.num_stage):
            y = self.linear_stages[i](y)
        #y = self.binarize(y)
        y = self.w2(y)
        #y = self.binarize(y)
        y = self.sigmoid(y)
        return y


class LookingNet_early_fusion_18(nn.Module):
    """
        Class definition for the combined Looking Model. Early fusion architecture with ResNet18 backbone. 
    """
    def __init__(self, PATH, PATH_look, device, fine_tune=True):
        """
        Args:
            PATH (str): Path to the pretrained ResNet18 heads model. Applicable only if fine-tune is enabled
            PATH_look (str): Path to the pretrained Looking joints model. Applicable only if fine-tune is enabled
            device (PyTorch device): PyTorch device
            fine_tune (bool, optional): Enable fine tune. Defaults to True.
        """
        super(LookingNet_early_fusion_18, self).__init__()
        self.backbone = ResNet18_head(device)
        if fine_tune:
            self.backbone.load_state_dict(torch.load(PATH, map_location=torch.device(device)))
            for m in self.backbone.net.parameters():
                m.requires_grad = False
            self.backbone.eval()

        self.looking_model = LookingModel(51)
        if fine_tune:
            self.looking_model.load_state_dict(torch.load(PATH_look, map_location=torch.device(device)))
            for m in self.looking_model.parameters():
                m.requires_grad = False
            self.looking_model.eval()



        self.encoder_head = nn.Sequential(
            nn.Linear(512, 256, bias=False),
            nn.BatchNorm1d(256),
            nn.Dropout(0.4),
            nn.ReLU(inplace=True)
        )

        
        # post processing

        self.looking_module = Looking_early_module()

    def forward(self, x):
        head, keypoint = x
        activation = {}
        def get_activation(name):
            def hook(model, input, output):
                activation[name] = output.detach().squeeze()
            return hook
        self.backbone.net.avgpool.register_forward_hook(get_activation('avgpool'))
        _ = self.backbone(head)

        layer_resnet = activation["avgpool"]
        encoded_head = self.encoder_head(layer_resnet)

        output_first_stage = self.looking_model.forward_first_stage(keypoint)
        y = self.looking_module.forward(output_first_stage+encoded_head)
        return y

class LookingNet_early_fusion_50(nn.Module):
    """
        Class definition for the combined Looking Model. Early fusion architecture with ResNext50 backbone. 
    """
    def __init__(self, PATH, PATH_look, device, fine_tune=True):
        """
        Args:
            PATH (str): Path to the pretrained ResNext50 heads model. Applicable only if fine-tune is enabled
            PATH_look (str): Path to the pretrained Looking joints model. Applicable only if fine-tune is enabled
            device (PyTorch device): PyTorch device
            fine_tune (bool, optional): Enable fine tune. Defaults to True.
        """
        super(LookingNet_early_fusion_50, self).__init__()
        self.backbone = ResNet50_head(device)
        if fine_tune:
            self.backbone.load_state_dict(torch.load(PATH))
            for m in self.backbone.net.parameters():
                m.requires_grad = False
            self.backbone.eval()

        self.looking_model = LookingModel(51)
        if fine_tune:
            self.looking_model.load_state_dict(torch.load(PATH_look, map_location=torch.device(device)))
            for m in self.looking_model.parameters():
                m.requires_grad = False
            self.looking_model.eval()



        self.encoder_head = nn.Sequential(
            nn.Linear(2048, 256, bias=False),
            nn.BatchNorm1d(256),
            nn.Dropout(0.4),
            nn.ReLU(inplace=True)
        )

        self.looking_module = Looking_early_module()

    def forward(self, x):
        head, keypoint = x
        activation = {}
        def get_activation(name):
            def hook(model, input, output):
                activation[name] = output.detach().squeeze()
            return hook
        self.backbone.net.avgpool.register_forward_hook(get_activation('avgpool'))
        _ = self.backbone(head)

        layer_resnet = activation["avgpool"]
        encoded_head = self.encoder_head(layer_resnet)

        output_first_stage = self.looking_model.forward_first_stage(keypoint)
        y = self.looking_module(output_first_stage+encoded_head)
        #y = self.looking_model.forward_second_stage(output_first_stage+encoded_head)
        return y

class EyesModel(nn.Module):
    """
        Class definition for the Eyes model. It consists of a bunch of FC layers that takes as an input the flattened
        representation of the eyes crops.
    """
    def __init__(self, device):
        """
        Args:
            device (PyTorch device): PyTorch device
        """
        super(EyesModel, self).__init__()

        self.encoder_eyes = nn.Sequential(
            nn.Flatten(),
            nn.Linear(900, 256, bias=False),
            nn.BatchNorm1d(256),
            nn.Dropout(0.4),
            nn.ReLU(inplace=True)
        ).to(device)
    
        self.fc = nn.Sequential(
            nn.Linear(256, 1),
            nn.Sigmoid()
        )
    def forward(self, x):
        encoded_eyes = self.encoder_eyes(x)
        return self.fc(encoded_eyes)



import os, errno
import argparse
import time

import logging

from .utils.network import *
from .utils.utils_predict import *

from PIL import Image, ImageFile

"""
PIFPAF is a command line tool and likes command line arguments, so for ease of use we use ArgParser to pass values to PIFPAF
"""
parser = argparse.ArgumentParser(prog='python3 predict', usage='%(prog)s [options] images', description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--batch_size', default=1, type=int, help='processing batch size')
parser.add_argument('--device', default='0', type=str, help='cuda device')
parser.add_argument('--long-edge', default=None, type=int, help='rescale the long side of the image (aspect ratio maintained)')
parser.add_argument('--loader-workers', default=None, type=int, help='number of workers for data loading')
parser.add_argument('--precise-rescaling', dest='fast_rescaling', default=True, action='store_false', help='use more exact image rescaling (requires scipy)')
parser.add_argument('--checkpoint_', default='shufflenetv2k30', type=str, help='backbone model to use')
parser.add_argument('--disable-cuda', default='False', action='store_false', help='disable CUDA')
parser.add_argument('--time', action='store_true', help='track comptutational time')
""""""

decoder.cli(parser)
logger.cli(parser)
network.Factory.cli(parser)
show.cli(parser)
visualizer.cli(parser)

args = parser.parse_args()

DOWNLOAD = None
INPUT_SIZE=51
FONT = cv2.FONT_HERSHEY_SIMPLEX

ImageFile.LOAD_TRUNCATED_IMAGES = True

log = logging.getLogger('openpifpaf.logger')
log.setLevel(logging.INFO)

class Predictor():
    """
        Class definition for the predictor.
        For user customization of the PIFPAF arguments, access is given to the PIFPAF args on 
        initialization of the class and are set again if changed from default defined above
    """
    def __init__(self, transparency=0.4, looking_threshold=0.5, mode='joints', device=args.device, pifpaf_ver='shufflenetv2k30', model_path='models/predictor',
                batch_size=args.batch_size, long_edge=args.long_edge, loader_workers=args.loader_workers, disable_cuda=args.disable_cuda):
        
        
        self.looking_threshold = looking_threshold
        self.transparency = transparency
        self.mode = mode

        #PIFPAF ARGS
        args.checkpoint = pifpaf_ver
        args.force_complete_pose = True
        args.batch_size=batch_size
        args.long_edge=long_edge
        args.load_workers=loader_workers
        args.disable_cuda=disable_cuda
        args.force_complete_pose = True
        #PIFPAF ARGS

        if device != 'cpu':
            use_cuda = torch.cuda.is_available()
            #print(torch.cuda.is_available())
            self.device = torch.device("cuda:{}".format(device) if use_cuda else "cpu")
        else:
            self.device = torch.device('cpu')
        args.device = self.device
        self.predictor_ = load_pifpaf(args)
        self.path_model = model_path
        try:
            os.makedirs(self.path_model)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        self.model = self.get_model().to(self.device)
        self.track_time = args.time
        if self.track_time:
            self.pifpaf_time = []
            self.inference_time = []
            self.total_time = []

        self.pred_labels=None
        self.pifpaf_outs=None
        self.output_image=None
        self.boxes=None
        self.keypoints=None

        log.info('OpenPifPaf version'+openpifpaf.__version__)
        log.info('PyTorch version'+torch.__version__)
        log.info(f'Device being used: {self.device}')

    
    def get_model(self):
        """Get Model
        Internal function to get the correcdt model to use
        Arguments:
        Returns:
            model
        """
        if self.mode == 'joints':
            model = LookingModel(INPUT_SIZE)
            if not os.path.isfile(os.path.join(self.path_model, 'LookingModel_LOOK+PIE.p')):
                """
                DOWNLOAD(LOOKING_MODEL, os.path.join(self.path_model, 'Looking_Model.zip'), quiet=False)
                with ZipFile(os.path.join(self.path_model, 'Looking_Model.zip'), 'r') as zipObj:
                    # Extract all the contents of zip file in current directory
                    zipObj.extractall()
                exit(0)"""
                raise NotImplementedError
            model.load_state_dict(torch.load(os.path.join(self.path_model, 'LookingModel_LOOK+PIE.p'), map_location=self.device))
            model.eval()
        else:
            model = AlexNet_head(self.device)
            if not os.path.isfile(os.path.join(self.path_model, 'AlexNet_LOOK.p')):
                """
                DOWNLOAD(LOOKING_MODEL, os.path.join(self.path_model, 'Looking_Model.zip'), quiet=False)
                with ZipFile(os.path.join(self.path_model, 'Looking_Model.zip'), 'r') as zipObj:
                    # Extract all the contents of zip file in current directory
                    zipObj.extractall()
                exit(0)"""
                raise NotImplementedError
            model.load_state_dict(torch.load(os.path.join(self.path_model, 'AlexNet_LOOK.p')))
            model.eval()
        return model

    def predict_look(self, boxes, keypoints, im_size, batch_wise=True) -> list:
        """Predict Look
        Internal function to predict looking using keypoints
        Arguments:
            bbox: list , the bounding boxes from PP prediction
            keypoints: list, the keypoints from PP prediction
            im_size: tuple , (w,h) image dimensions
            batch_wise: bool
        Returns:
            out_labels: list
        """
        label_look = []
        final_keypoints = []
        if batch_wise:
            if len(boxes) != 0:
                for i in range(len(boxes)):
                    kps = keypoints[i]
                    kps_final = np.array([kps[0], kps[1], kps[2]]).flatten().tolist()
                    X, Y = kps_final[:17], kps_final[17:34]
                    X, Y = normalize_by_image_(X, Y, im_size)
                    #X, Y = normalize(X, Y, divide=True, height_=False)
                    kps_final_normalized = np.array([X, Y, kps_final[34:]]).flatten().tolist()
                    final_keypoints.append(kps_final_normalized)
                tensor_kps = torch.Tensor([final_keypoints]).to(self.device)
                if self.track_time:
                    start = time.time()
                    out_labels = self.model(tensor_kps.squeeze(0)).detach().cpu().numpy().reshape(-1)
                    end = time.time()
                    self.inference_time.append(end-start)
                else:
                    out_labels = self.model(tensor_kps.squeeze(0)).detach().cpu().numpy().reshape(-1)
            else:
                out_labels = []
        else:
            if len(boxes) != 0:
                for i in range(len(boxes)):
                    kps = keypoints[i]
                    kps_final = np.array([kps[0], kps[1], kps[2]]).flatten().tolist()
                    X, Y = kps_final[:17], kps_final[17:34]
                    X, Y = normalize_by_image_(X, Y, im_size)
                    #X, Y = normalize(X, Y, divide=True, height_=False)
                    kps_final_normalized = np.array([X, Y, kps_final[34:]]).flatten().tolist()
                    #final_keypoints.append(kps_final_normalized)
                    tensor_kps = torch.Tensor(kps_final_normalized).to(self.device)
                    if self.track_time:
                        start = time.time()
                        out_labels = self.model(tensor_kps.unsqueeze(0)).detach().cpu().numpy().reshape(-1)
                        end = time.time()
                        self.inference_time.append(end-start)
                    else:
                        out_labels = self.model(tensor_kps.unsqueeze(0)).detach().cpu().numpy().reshape(-1)
            else:
                out_labels = []
        return out_labels
    
    def predict_look_alexnet(self, boxes, image, batch_wise=True) -> list:
        """Predict Look AlexNet
        Internal function to predict looking using AlexNet
        Arguments:
            bbox: list , the bounding boxes from PP prediction
            image: internal format image
            batch_wise: bool
        Returns:
            out_labels: list
        """
        out_labels = []
        data_transform = transforms.Compose([
                        SquarePad(),
                        transforms.Resize((227,227)),
                        transforms.ToTensor(),
                        transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                            std=[0.229, 0.224, 0.225])])
        if len(boxes) != 0:
            if batch_wise:
                heads = []
                for i in range(len(boxes)):
                    bbox = boxes[i]
                    x1, y1, x2, y2, _ = bbox
                    w, h = abs(x2-x1), abs(y2-y1)
                    head_image = Image.fromarray(np.array(image)[int(y1):int(y1+(h/3)), int(x1):int(x2), :])
                    head_tensor = data_transform(head_image)
                    heads.append(head_tensor.detach().cpu().numpy())
                if self.track_time:
                    start = time.time()
                    out_labels = self.model(torch.Tensor([heads]).squeeze(0).to(self.device)).detach().cpu().numpy().reshape(-1)
                    end = time.time()
                    self.inference_time.append(end-start)
            else:
                out_labels = []
                for i in range(len(boxes)):
                    bbox = boxes[i]
                    x1, y1, x2, y2, _ = bbox
                    w, h = abs(x2-x1), abs(y2-y1)
                    head_image = Image.fromarray(np.array(image)[int(y1):int(y1+(h/3)), int(x1):int(x2), :])
                    head_tensor = data_transform(head_image)
                    #heads.append(head_tensor.detach().cpu().numpy())
                    if self.track_time:
                        start = time.time()
                        looking_label = self.model(torch.Tensor(head_tensor).unsqueeze(0).to(self.device)).detach().cpu().numpy().reshape(-1)[0]
                        end = time.time()
                        self.inference_time.append(end-start)
                    else:
                        looking_label = self.model(torch.Tensor(head_tensor).unsqueeze(0).to(self.device)).detach().cpu().numpy().reshape(-1)[0]
                    out_labels.append(looking_label)
                #if self.track_time:
                #    out_labels = self.model(torch.Tensor([heads]).squeeze(0).to(self.device)).detach().cpu().numpy().reshape(-1)
        else:
            out_labels = []
        return out_labels

    def render_image(self, image, bbox, keypoints, pred_labels) -> np.ndarray:
        """Render Image
        Internal function that takes the iternal image format and predictions and outputs 
        the image with overlayed predictions
        Arguments:
            image: internal format image
            bbox: list , the bounding boxes from PP prediction
            keypoints: list , the keypoints from PP prediction
            pred_labels: list , the predictions of looking attention
        Returns:
            open_cv_image: np.ndarray , the opencv format image
        """
        open_cv_image = np.array(image) 
        open_cv_image = open_cv_image[:, :, ::-1].copy()
        
        scale = 0.007
        imageWidth, imageHeight, _ = open_cv_image.shape
        font_scale = min(imageWidth,imageHeight)/(10/scale)

        mask = np.zeros(open_cv_image.shape, dtype=np.uint8)
        for i, label in enumerate(pred_labels):
            if label > self.looking_threshold:
                color = (0,255,0)
            else:
                color = (255,0,0)
            mask = draw_skeleton(mask, keypoints[i], color)
        mask = cv2.erode(mask,(7,7),iterations = 1)
        mask = cv2.GaussianBlur(mask,(3,3),0)
        open_cv_image = cv2.addWeighted(open_cv_image, 1, mask, self.transparency, 1.0)
        open_cv_image = cv2.cvtColor(open_cv_image, cv2.COLOR_BGR2RGB)

        return open_cv_image

    def predict(self, image: np.ndarray):
        """Prediction Function
        Arguments:
            image: a RGB image in the form of a numpy array(OpenCV format)
        Returns:
            None
        """
        self.input_image = image

        if(image is None or image.ndim!=3):
            log.error('ERROR: Image is none or of wrong format')
            exit()
        img_width, img_height, c = image.shape
        
        try:
            cpu_image = Image.fromarray(image)

            pp_predictions, pp_anns, pp_meta = self.predictor_.numpy_image(image)

            pifpaf_outs = {
                'json_data' : [ann.json_data() for ann in pp_predictions],
                'image' : cpu_image
            }

            im_size = (cpu_image.size[0], cpu_image.size[1])
            self.boxes, self.keypoints = preprocess_pifpaf(pifpaf_outs['json_data'], im_size, enlarge_boxes=False)
            if self.mode == 'joints':
                self.pred_labels = self.predict_look(self.boxes, self.keypoints, im_size)
            else:
                self.pred_labels = self.predict_look_alexnet(self.boxes, cpu_image)

            self.output_image = self.render_image(pifpaf_outs['image'], self.boxes, self.keypoints, self.pred_labels)

        except Exception as err:
            log.error(f'{err}, {type(err)}')

    def get_pred_labels(self) -> list:
        """Prediction Labels
        Returns:
            list: pred_labels
        """
        return self.pred_labels
    def get_pifpaf_output(self) -> dict:
        """PIFPAF Output Dictionary
        Returns:
            dict: pifpaf_outs
        """
        return self.pifpaf_outs
    def get_output_image(self) -> np.ndarray:
        """OpenCV Output Image
        Returns:
            np.ndarray: output_image
        """
        return self.output_image
    def get_bounding_boxes(self) -> list:
        """Bounding Boxes
        Returns:
            list: boxes
        """
        return self.boxes
    def get_keypoints(self) -> list:
        """Keypoints
        Returns:
            list: keypoints
        """
        return self.keypoints


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="python-look",
    version="0.1.0",
	license="MIT",
    author="Carson Wittwer",
    author_email="wittwer.carson@googlemail.com",
    description="Simple wrapper code to run inference on images using the model and code from the Looking Repo from VITA-EPFL.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cwittwer/look",
    packages=find_packages(exclude=("tests", "requirements.txt",)),
	include_package_data=True,
	install_requires=[
    "tqdm",
    "numpy",
    "opencv_python==4.7.0.72",
    "seaborn",
    "matplotlib",
    "torch",
    "torchvision",
    "openpifpaf==0.13.11",
    "scikit_image==0.19.3",
    "Pillow",
    "scikit_learn"
	],
    classifiers=[
      "Development Status :: 4 - Beta",
		  "Intended Audience :: Developers",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
      "Programming Language :: Python :: 3.6",
		  "Programming Language :: Python :: 3.7",
		  "Programming Language :: Python :: 3.8",
		  "Programming Language :: Python :: 3.9"
    ],
    python_requires='>=3.6, <3.10'
)




import cv2
import openpifpaf
import os

import openpifpaf
import numpy as np
from datetime import datetime
from openpifpaf import datasets
from openpifpaf import decoder, network, visualizer, show, logger, Predictor
from openpifpaf.predict import out_name

"""COCO_PERSON_SKELETON = [
        [16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13],
    [6, 7], [6, 8], [7, 9], [8, 10], [9, 11], [2, 3], [1, 2], [1, 3],
    [2, 4], [3, 5], [4, 6], [5, 7]]
"""
COCO_PERSON_SKELETON = [
        [16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13],
    [6, 7], [6, 8], [7, 9], [8, 10], [9, 11]]
COCO_HEAD = [
    [3,4]
]
def convert(data):
    """X = []
    Y = []
    C = []
    A = []
    i = 0
    print(data)
    while i < 51:
        X.append(data[i])
        Y.append(data[i+1])
        C.append(data[i+2])
        i += 3
    """
    X, Y, C = data[0], data[1], data[2]
    A = np.array([X, Y, C]).flatten().tolist()
    return X, Y, C, A

def normalize_by_image_(X, Y, image_size):
    """
        Normalize the image according to the paper.
        Args:
            - X: array of X positions of the keypoints
            - Y: array of Y positions of the keypoints
            - Image: Image array
        Returns:
            returns the normalized arrays
    """
    
    image_width, image_height = image_size
    
    center_p = (int((X[11] + X[12]) / 2), int((Y[11] + Y[12]) / 2))
    X_new = np.array(X)/image_width
    Y_new = np.array(Y)-center_p[1]


    width = abs(np.max(X) - np.min(X))
    height = abs(np.max(Y) - np.min(Y))

    X_new = X_new + ((np.array(X)-center_p[0])/width)
    Y_new /= height
    
    return X_new, Y_new


def prepare_pif_kps(kps_in):
    """Convert from a list of 51 to a list of 3, 17"""

    assert len(kps_in) % 3 == 0, "keypoints expected as a multiple of 3"
    xxs = kps_in[0:][::3]
    yys = kps_in[1:][::3]  # from offset 1 every 3
    ccs = kps_in[2:][::3]

    return [xxs, yys, ccs]


def preprocess_pifpaf(annotations, im_size=None, enlarge_boxes=True, min_conf=0.):
    """
    Preprocess pif annotations:
    1. enlarge the box of 10%
    2. Constraint it inside the image (if image_size provided)
    """

    boxes = []
    keypoints = []
    enlarge = 1 if enlarge_boxes else 2  # Avoid enlarge boxes for social distancing

    for dic in annotations:
        kps = prepare_pif_kps(dic['keypoints'])
        box = dic['bbox']
        try:
            conf = dic['score']
            # Enlarge boxes
            delta_h = (box[3]) / (10 * enlarge)
            delta_w = (box[2]) / (5 * enlarge)
            # from width height to corners
            box[2] += box[0]
            box[3] += box[1]

        except KeyError:
            all_confs = np.array(kps[2])
            score_weights = np.ones(17)
            score_weights[:3] = 3.0
            score_weights[5:] = 0.1
            # conf = np.sum(score_weights * np.sort(all_confs)[::-1])
            conf = float(np.mean(all_confs))
            # Add 15% for y and 20% for x
            delta_h = (box[3] - box[1]) / (7 * enlarge)
            delta_w = (box[2] - box[0]) / (3.5 * enlarge)
            assert delta_h > -5 and delta_w > -5, "Bounding box <=0"

        box[0] -= delta_w
        box[1] -= delta_h
        box[2] += delta_w
        box[3] += delta_h

        # Put the box inside the image
        if im_size is not None:
            box[0] = max(0, box[0])
            box[1] = max(0, box[1])
            box[2] = min(box[2], im_size[0])
            box[3] = min(box[3], im_size[1])

        if conf >= min_conf:
            box.append(conf)
            boxes.append(box)
            keypoints.append(kps)

    return boxes, keypoints

def filecreation(dirname):
    mydir = os.path.join(dirname,'output_'+datetime.now().strftime('%Y-%m-%d_%H-%M-%S'))
    try:
        pass
        #os.makedirs(mydir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise  # This was not a "directory exist" error..
    return mydir

def draw_skeleton(img, kps, color, skeleton=COCO_PERSON_SKELETON):
    X, Y, C, _ = convert(kps)
    linewidth = 4
    height = abs(Y[0]-Y[-1])
    for ci, connection in enumerate(np.array(skeleton) - 1):
        #print(connection)
        c1, c2 = connection
        #print(X[c1])
        img = cv2.line(img,(int(X[c2]),int(Y[c2])),(int(X[c1]),int(Y[c1])), color,linewidth)
    head = COCO_HEAD[0]
    c1, c2 = head
    #radius = abs(int(X[c1])- int(X[c2]))
    radius = int(0.09*height)
    center = int((X[c1]+X[c2])/2), int((Y[c1]+Y[c2])/2)
    img = cv2.circle(img, center, radius, color, -1)
    img = cv2.circle(img, center, radius, (255, 255, 255), 2)
    return img

def run_and_kps(img, kps, label):
    blk = np.zeros(img.shape, np.uint8)
    X, Y, C, _ = convert(kps)
    if label > 0.5:
        color = (0, 255, 0)
    else:
        color = (0,0,255)
    for i in range(len(Y)):
        blk =cv2.circle(blk, (int(X[i]), int(Y[i])), 1, color, 2)
    #img = cv2.addWeighted(img, 1.0, blk, 0.55, 1)
    return blk

#def load_pifpaf(args):
#    args.figure_width = 10
#    args.dpi_factor = 1.0
#    args.keypoint_threshold_rel = 0.0
#    args.instance_threshold = 0.2
#    args.keypoint_threshold = 0
#    args.force_complete_pose = True

#    decoder.configure(args)
#    network.Factory.configure(args)
#    show.configure(args)
#    visualizer.configure(args)

    #processor, net = processor_factory(args)
    #preprocess = preprocess_factory(args)
    #return net, processor, preprocess

def load_pifpaf(args):
    pifpaf_model = args.checkpoint_

    args.figure_width = 10
    args.dpi_factor = 1.0
    args.batch_size = 1


    args.keypoint_threshold_rel = 0.0
    args.instance_threshold = 0.2
    args.keypoint_threshold = 0

    args.force_complete_pose = True


    # Configure
    decoder.configure(args)
    network.Factory.configure(args)
    Predictor.configure(args)
    show.configure(args)
    visualizer.configure(args)

    return Predictor(checkpoint=pifpaf_model)